/*
 * lxdash-i.l.c
 * Copyright (C) 2011-2018, 2022, 2024, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <d_fset.h>
#include <lxdash-inter.h>
#include <lxdash-drive.h>
#include <lxdash-types.h>

#define dxdash(list) \
    ((struct dxdash_type *) (list))
#define fxroll(roll) \
    ((struct fxroll_type *) (roll))
#define lxdash(list) \
    ((struct lxdash_type *) (list))

typedef struct rule_type {
    X1f4_E4_C_MODE tail;
    int over;
    struct either_type either;
} rule_type;

static int rule_dash(void *, void *);

static int
rule_dash(void *line, void *data)
{
    int status;
    struct rule_type *rule_data;

    rule_data = line;

    if (((struct effect_type *) data)->mode < rule_data->tail) {
	status = _libx1f4i0_lxdash_rule_data(&rule_data->either, data);
    } else {
	rule_data->over = 1;
	status = 1;
    }

    return status;
}


int
_libx1f4i0_lxdash_wish_line(struct lxrail_type *lxrail_data,
			    struct affect_type *affect_data, void *dxdash,
			    void *fxroll, unsigned *miss)
{
    int status;
    struct rule_type rule;

    rule.either.excase_data = lock_excase(dxdash(dxdash));
    rule.either.affect_data = affect_data;
    rule.either.lxtext_data = lxdash(lxrail_data)->link_f.text;

    rule.either.roll = 0;

    if (fxroll(fxroll)->tail < fxroll(fxroll)->head) {
	status = _libx1f4i0_fxpd_lxfset
	    (dxdash(dxdash)->dash, &fxroll(fxroll)->head, &rule.either,
	     _libx1f4i0_lxdash_rule_data);
	if (status) {
	} else {
#if 0
	    status = (infer ? _libx1f4i0_lxpd_lxfset : _libx1f4i0_llpd_lxfset)
#else
	    status = _libx1f4i0_llpd_lxfset
#endif				/* 0 */
		(dxdash(dxdash)->dash, &fxroll(fxroll)->tail, &rule.either,
		 _libx1f4i0_lxdash_rule_data);
	}
    } else {
	rule.over = 0;
	rule.tail = fxroll(fxroll)->tail;
	status = _libx1f4i0_fxpd_lxfset
	    (dxdash(dxdash)->dash, &fxroll(fxroll)->head, &rule, rule_dash);
	if (rule.over) {
	    status = 0;
	}
    }

    if (rule.either.roll) {
	*miss = rule.either.affect_data - affect_data;
    }

    return status;
}
