/*
 * lxvery-i.h.c
 * Copyright (C) 2011-2018, 2021, 2022, 2024, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <lxvery-config.h>

#include <d_fset.h>
#include <lxvery-defs.h>
#include <lxvery-drive.h>
#include <lxvery-types.h>

#define __SINGLE__			1

#define dxvery(list) \
    ((struct dxvery_type *) (list))
#define fxroll(roll) \
    ((struct fxroll_type *) (roll))

#define rule(___s) \
    ((struct rule_type *) (___s))

typedef struct rule_type {
    X1f4_E4_C_MODE mode;
    struct fydata_type *fydata_data;
} rule_type;

static int rule_dash(void *, void *);
static int slip_cast(void *, void *);

static int
rule_dash(void *line, void *data)
{
    struct fydata_type **fydata;

    fydata = line;

    l_MODE(*(X1f4_E4_C_MODE *) data - VERY_BASE, *fydata);

    (*fydata)++;

    return 0;
}


static int
slip_cast(void *rule, void *data)
{
    F_TYPE mode;
    int delete;

    mode = *(X1f4_E4_C_MODE *) data;
    if (mode < rule(rule)->mode) {
	l_MODE(*(X1f4_E4_C_MODE *) data - VERY_BASE, rule(rule)->fydata_data);

	rule(rule)->fydata_data++;

	delete = 0;
    } else {
	delete = 1;
    }

    return delete;
}


int
_libx1f4i0_lxvery_vamp_line(struct lxvery_type *lxvery_data,
			    struct fydata_type *fydata_data, void *dxvery,
			    void *fxroll)
{
    F_TYPE head, tail;

    head = fxroll(fxroll)->head;
#if __VERY_PSET__
    head += ~(X1f4_E4_C_BILL) 0 >> 1;
#endif				/* __VERY_PSET__ */

    tail = fxroll(fxroll)->tail;
#if __VERY_PSET__
    tail += ~(X1f4_E4_C_BILL) 0 >> 1;
#endif				/* __VERY_PSET__ */

#if __SINGLE__
    if (tail < head) {
	void *line;

	line = fydata_data;
	_libx1f4i0_fxpd_lxfset(dxvery(dxvery)->dash, &head, &line, rule_dash);
	_libx1f4i0_llpd_lxfset(dxvery(dxvery)->dash, &tail, &line, rule_dash);
    } else {
	struct rule_type rule;

	rule.fydata_data = fydata_data;
	rule.mode = tail;
	_libx1f4i0_fxpd_lxfset(dxvery(dxvery)->dash, &head, &rule, slip_cast);
    }
#else
#endif				/* __SINGLE__ */

    return 0;
}
